# Copyright 2021-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from typing import List

import math
from mindspore.common.initializer import HeUniform
from mindspore.common.initializer import initializer
from mindspore.common.initializer import Uniform
from mindspore.common.initializer import _calculate_fan_in_and_fan_out
from mindspore import dtype as mstype
from mindspore.ops import operations as P
from mindspore.ops import functional as F
from mindspore import Parameter
import mindspore.nn as nn
import mindspore.ops as ops


class HeDense(nn.Dense):
    def __init__(self, in_channels, out_channels, has_bias=True,
                 activation=None):
        super().__init__(in_channels, out_channels, weight_init='normal',
                         bias_init='zeros', has_bias=has_bias,
                         activation=activation)
        self.reset_parameters()

    def reset_parameters(self):
        self.weight.set_data(
            initializer(HeUniform(math.sqrt(5)), self.weight.shape))
        if self.has_bias is not None:
            fan_in, _ = _calculate_fan_in_and_fan_out(self.weight.shape)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            self.bias.set_data(
                initializer(Uniform(bound), [self.out_channels]))


class Gaussian(nn.Cell):
    def __init__(self, num_group_id, latent_dim):
        super(Gaussian, self).__init__()
        self.num_group_id = num_group_id

        """defined ops"""
        self.expand_dims = P.ExpandDims()

        init_tensor = initializer('zeros',
                                  [self.num_group_id, latent_dim])

        self.params = Parameter(init_tensor,
                                name="mean_y",
                                requires_grad=True)

    def construct(self, input_layer):
        latent_layer = input_layer
        latent_layer = self.expand_dims(latent_layer, 1)

        return latent_layer - self.expand_dims(self.params, 0)


class VAE(nn.Cell):

    def __init__(self,
                 input_shape: int,
                 latent_dim: int,
                 num_group_id: int,
                 dropout_p=0.1,
                 hidden_dims: List = None,
                 class_dims: List = None) -> None:
        super(VAE, self).__init__()

        if type(input_shape) != int:
            raise ValueError("input_shape type is not right")

        if type(latent_dim) != int:
            raise ValueError("latent_dim type is not right")

        if type(num_group_id) != int:
            raise ValueError("num_group_id type is not right")

        """ Init parameter """
        self.input_shape = input_shape
        self.latent_dim = latent_dim
        self.num_group_id = num_group_id
        self.hidden_dims = hidden_dims
        self.class_dims = class_dims
        self.dropout_p = dropout_p

        """defined some ops in here"""
        self.exp = P.Exp()
        self.shape = P.Shape()

        self.flatten = nn.Flatten()

        self.encoder = None
        self.decoder = None
        self.classifier = None
        self.final_layer = None
        self.fc_mu = None
        self.fc_var = None

        """Estimated parameter"""
        self.fc_mu = HeDense(self.hidden_dims[-1], self.latent_dim)
        self.fc_var = HeDense(self.hidden_dims[-1], self.latent_dim)

        """Link latent layer ---> decoder layer"""
        self.decoder_input = HeDense(self.latent_dim, hidden_dims[-1])

        """Use to Estimated mean_y parameter"""
        self.gaussian_layer = Gaussian(self.num_group_id, self.latent_dim)

        """Build network component"""
        self.build_encoder()
        self.build_decoder()
        self.build_classifier()

    def build_encoder(self):

        if self.hidden_dims is None:
            self.hidden_dims = [32]

        encoder_block = self._build_block(self.input_shape,
                                          self.hidden_dims)

        self.encoder = nn.SequentialCell(*encoder_block)

    def build_decoder(self):
        # Build Decoder
        self.hidden_dims.reverse()

        decoder_block = self._build_block(self.hidden_dims[0],
                                          self.hidden_dims)

        self.decoder = nn.SequentialCell(*decoder_block)
        self.build_final_layer(self.hidden_dims[-1])

    def build_classifier(self):

        modules = []
        if self.class_dims is None:
            self.class_dims = [32]

        layer_number = self.latent_dim

        for c_dim in self.class_dims:
            modules.append(
                HeDense(layer_number, c_dim))
            modules.append(nn.BatchNorm1d(c_dim))
            modules.append(nn.ReLU())
            modules.append(nn.Dropout(keep_prob=(1 - self.dropout_p)))

            layer_number = c_dim

        modules.append(
            HeDense(self.class_dims[-1], self.num_group_id)
        )

        # Build classifier
        self.classifier = nn.SequentialCell(*modules)

    def build_final_layer(self, decode_layer_number):
        # build
        self.final_layer = nn.SequentialCell(
            HeDense(decode_layer_number, 256),
            nn.BatchNorm1d(256),
            nn.ReLU(),
            HeDense(256, self.input_shape)
        )

    def encode(self, input_layer):
        """Handle Encode Network"""
        result = self.encoder(input_layer)
        result = self.flatten(result)

        mu = self.fc_mu(result)
        log_var = self.fc_var(result)

        return [mu, log_var]

    def decode(self, z_layer):
        """Handle Decode Network"""
        result = self.decoder_input(z_layer)
        result = self.decoder(result)
        result = self.final_layer(result)

        return result

    def classify(self, z):
        classify_layer = self.classifier(z)

        pred_hat = nn.Softmax(1)(classify_layer)

        return pred_hat

    def construct(self, input_layer, z_std):

        x = self.flatten(input_layer)

        mu, log_var = self.encode(x)

        std = self.exp(0.5 * log_var)
        z = z_std * std + mu

        z_prior_mean = self.gaussian_layer(z)

        y = self.classify(z)
        x_recon = self.decode(z)

        return [x_recon, x, z_prior_mean, log_var, y]

    """
        Utils Function use in Model
    """

    @property
    def name(self):
        return 'VAE'

    def _build_block(self, start_shape, sequence_dim):
        modules = []

        num_layer_input = start_shape
        for h_dim in sequence_dim:
            modules.append(
                HeDense(num_layer_input, h_dim))
            modules.append(nn.BatchNorm1d(h_dim))
            modules.append(nn.ReLU())
            modules.append(nn.Dropout(keep_prob=(1 - self.dropout_p)))

            num_layer_input = h_dim

        return modules


class VaeLoss(nn.Cell):
    def __init__(self):
        """init class"""
        super(VaeLoss, self).__init__()

        self.expand_dims = P.ExpandDims()
        self.square = P.Square()
        self.log = P.Log()
        self.reduce_sum = P.ReduceSum(keep_dims=False)
        pass

    def construct(self, recon, base, latent_prior_mean, log_var, y):
        log_var = self.expand_dims(log_var, 1)

        """ first loss """
        recons_loss = 0.5 * self.square((recon - base)).mean(0)

        """ second loss """
        latent_mu_square = self.square(latent_prior_mean)
        kl_loss = - 0.5 * (log_var - latent_mu_square)
        matmul_kl = ops.matmul(self.expand_dims(y, 1),
                               kl_loss)
        kld_loss = matmul_kl.mean(0)

        """ third """
        log_y = self.log(y + 1e-8)
        cat_loss = (y * log_y).mean(0)

        lamb = 1
        final_loss = lamb * self.reduce_sum(recons_loss) + \
                     self.reduce_sum(kld_loss) + \
                     self.reduce_sum(cat_loss)

        return final_loss


class VaeWithLossCell(nn.Cell):
    def __init__(self, backbone, loss_fn):
        super(VaeWithLossCell, self).__init__(auto_prefix=False)
        self._backbone = backbone
        self._loss_fn = loss_fn

    def construct(self, data, z):
        base = self._backbone(data, z)
        return self._loss_fn(base[0], base[1], base[2], base[3], base[4])


class VaeTrainOneStepCell(nn.Cell):
    def __init__(self, network, optimizer, sens=1.0):
        super(VaeTrainOneStepCell, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_grad()
        self.optimizer = optimizer
        self.weights = self.optimizer.parameters
        self.grad = ops.GradOperation(get_by_list=True, sens_param=True)
        self.sens = sens

    def construct(self, *inputs):
        loss = self.network(*inputs)
        sens = F.fill(loss.dtype, loss.shape, self.sens)
        grads = self.grad(self.network, self.weights)(*inputs, sens)
        loss = F.depend(loss, self.optimizer(grads))
        return loss


class VaeWithEvalCell(nn.Cell):
    def __init__(self, network, loss_fn, add_cast_fp32=False):
        super(VaeWithEvalCell, self).__init__(auto_prefix=False)
        self._network = network
        self._loss_fn = loss_fn
        self.add_cast_fp32 = add_cast_fp32

    def construct(self, data, label, z_tensor):
        outputs = self._network(data, z_tensor)
        if self.add_cast_fp32:
            label = F.mixed_precision_cast(mstype.float32, label)
            outputs = F.cast(outputs, mstype.float32)
        return outputs, label
