# Copyright 2021-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import argparse
import os
import logging
import ast
import sys

import numpy as np

from mindspore import nn, Tensor, context, set_seed, load_checkpoint, load_param_into_net

from Horizontal_FL.device.nets.vae_net.vae_ms import VAE
from Horizontal_FL.device.nets.vae_net.vae_ms import VaeLoss
from Horizontal_FL.device.nets.vae_net.vae_ms import VaeWithLossCell
from Horizontal_FL.device.nets.vae_net.vae_ms import VaeTrainOneStepCell
from Horizontal_FL.common.utils import load_config


def rand_z(batch_size=1, latent_dim=64, seed=None):
    if seed is not None:
        np.random.seed(seed)
    z_stn = np.random.standard_normal(size=(batch_size, latent_dim)).astype(np.float32)
    z_stn = Tensor(z_stn)
    return z_stn


def cloud_task(net, init_checkpoint_dir="none", batch_size=1, input_shape=1039, latent_dim=64):
    crit = VaeLoss()
    net_with_criterion = VaeWithLossCell(net, crit)
    opt = nn.Adam(params=net.trainable_params())
    train_net = VaeTrainOneStepCell(net_with_criterion, opt)
    if os.path.exists(init_checkpoint_dir):
        for checkpoint_file in os.listdir(checkpoint_dir):
            print("Checkpoint file is {}".format(checkpoint_file))
            file_prefix = fl_name + "_iteration_"
            if checkpoint_file.startswith(file_prefix):
                checkpoint_file_path = os.path.join(checkpoint_dir, checkpoint_file)
                param_dict = load_checkpoint(checkpoint_file_path)
                load_param_into_net(train_net, param_dict)
                print("Read checkpoint file success, checkpoint_file_path is {}".format(checkpoint_file_path))
                break
        else:
            logging.info("Checkpoint dir {} is not exists".format(checkpoint_dir))
    sys.stdout.flush()
    train_net.set_train()
    fake_data = Tensor(np.ones(shape=(batch_size, input_shape), dtype=np.float32))
    z_std = rand_z(batch_size, latent_dim, seed=0)
    train_net(fake_data, z_std)


def init_model(config):
    input_shape = config['vae_params']['input_shape']
    latent_dim = config['vae_params']['latent_dim']
    hidden_dims = config['vae_params']['hidden_dim']
    class_dims = config['vae_params']['class_dim']
    dropout_p = config['vae_params']['dropout_p']
    num_group_id = config['vae_params']['num_group_id']

    net = VAE(input_shape, latent_dim,
              num_group_id, dropout_p,
              hidden_dims=hidden_dims,
              class_dims=class_dims)

    return net


def parse_args():
    """
    parse args
    """
    parser = argparse.ArgumentParser(description="vae_task")
    parser.add_argument("--enable_fl", type=bool, default=False)
    parser.add_argument("--device_target", type=str, default="GPU")
    parser.add_argument("--server_mode", type=str, default="FEDERATED_LEARNING")
    parser.add_argument("--ms_role", type=str, default="MS_SERVER")
    parser.add_argument("--worker_num", type=int, default=0)
    parser.add_argument("--server_num", type=int, default=1)
    parser.add_argument("--scheduler_ip", type=str, default="127.0.0.1")
    parser.add_argument("--scheduler_port", type=int, default=4444)
    parser.add_argument("--fl_server_port", type=int, default=5666)
    parser.add_argument("--start_fl_job_threshold", type=int, default=10)
    parser.add_argument("--start_fl_job_time_window", type=int, default=30000000)
    parser.add_argument("--update_model_ratio", type=float, default=1.0)
    parser.add_argument("--update_model_time_window", type=int, default=30000000)
    parser.add_argument("--fl_name", type=str, default="VAE")
    parser.add_argument("--fl_iteration_num", type=int, default=500000)
    parser.add_argument("--client_epoch_num", type=int, default=1)
    parser.add_argument("--client_batch_size", type=int, default=100)
    parser.add_argument("--client_learning_rate", type=float, default=0.001)
    parser.add_argument("--encrypt_type", type=str, default="NOT_ENCRYPT",
                        choices=["NOT_ENCRYPT", "DP_ENCRYPT", "SIGNDS", "PW_ENCRYPT"])
    parser.add_argument("--scheduler_manage_port", type=int, default=12202)
    parser.add_argument("--config_file_path", type=str, default="{}/config.json".format(os.getcwd()))
    parser.add_argument("--enable_ssl", type=ast.literal_eval, default=False)
    # parameters for encrypt_type='DP_ENCRYPT'
    parser.add_argument("--dp_eps", type=float, default=50.0)
    parser.add_argument("--dp_delta", type=float, default=0.01)  # usually equals 1/start_fl_job_threshold
    parser.add_argument("--dp_norm_clip", type=float, default=0.05)
    # parameters for encrypt_type="SIGNDS"
    parser.add_argument("--sign_k", type=float, default=0.01)
    parser.add_argument("--sign_eps", type=float, default=100)
    parser.add_argument("--sign_thr_ratio", type=float, default=0.6)
    parser.add_argument("--sign_global_lr", type=float, default=2)
    parser.add_argument("--sign_dim_out", type=int, default=0)
    # parameters for encrypt_type="PW_ENCRYPT"
    parser.add_argument("--share_secrets_ratio", type=float, default=1.0)
    parser.add_argument("--cipher_time_window", type=int, default=3000)
    parser.add_argument("--reconstruct_secrets_threshold", type=int, default=2)
    parser.add_argument("--root_first_ca_path", type=str, default="")
    parser.add_argument("--root_second_ca_path", type=str, default="")
    parser.add_argument("--pki_verify", type=ast.literal_eval, default=False)
    parser.add_argument("--equip_crl_path", type=str, default="")
    # Time difference limit for Anti-Replay Attacks. The default value is 60 seconds.
    parser.add_argument("--replay_attack_time_diff", type=int, default=60000)
    parser.add_argument("--http_url_prefix", type=str, default="")
    parser.add_argument("--global_iteration_time_window", type=int, default=3600000)
    # init checkpoint dir
    parser.add_argument("--checkpoint_dir", type=str, default="")
    # vae args
    parser.add_argument("--server_config_path", type=str, default="./Horizontal_FL/conf/server_config.yaml")

    return parser.parse_args()


if __name__ == "__main__":
    set_seed(0)
    args = parse_args()
    enable_fl = args.enable_fl
    device_target = args.device_target
    server_mode = args.server_mode
    ms_role = args.ms_role
    worker_num = args.worker_num
    server_num = args.server_num
    scheduler_ip = args.scheduler_ip
    scheduler_port = args.scheduler_port
    fl_server_port = args.fl_server_port
    start_fl_job_threshold = args.start_fl_job_threshold
    start_fl_job_time_window = args.start_fl_job_time_window
    update_model_ratio = args.update_model_ratio
    update_model_time_window = args.update_model_time_window
    fl_name = args.fl_name
    fl_iteration_num = args.fl_iteration_num
    client_epoch_num = args.client_epoch_num
    client_batch_size = args.client_batch_size
    client_learning_rate = args.client_learning_rate
    scheduler_manage_port = args.scheduler_manage_port
    config_file_path = args.config_file_path
    enable_ssl = args.enable_ssl
    encrypt_type = args.encrypt_type
    dp_eps = args.dp_eps
    dp_delta = args.dp_delta
    dp_norm_clip = args.dp_norm_clip
    sign_k = args.sign_k
    sign_eps = args.sign_eps
    sign_thr_ratio = args.sign_thr_ratio
    sign_global_lr = args.sign_global_lr
    sign_dim_out = args.sign_dim_out
    share_secrets_ratio = args.share_secrets_ratio
    cipher_time_window = args.cipher_time_window
    reconstruct_secrets_threshold = args.reconstruct_secrets_threshold
    root_first_ca_path = args.root_first_ca_path
    root_second_ca_path = args.root_second_ca_path
    pki_verify = args.pki_verify
    equip_crl_path = args.equip_crl_path
    replay_attack_time_diff = args.replay_attack_time_diff
    http_url_prefix = args.http_url_prefix
    global_iteration_time_window = args.global_iteration_time_window

    # init ckpt
    checkpoint_dir = args.checkpoint_dir

    # vae args
    server_config_path = args.server_config_path

    logging.basicConfig(
        level=logging.DEBUG,
        format="[%(levelname)s] %(asctime)s [%(filename)s:%(lineno)s] %(message)s",
    )
    ctx = {
        "enable_fl": enable_fl,
        "server_mode": server_mode,
        "ms_role": ms_role,
        "worker_num": worker_num,
        "server_num": server_num,
        "scheduler_ip": scheduler_ip,
        "scheduler_port": scheduler_port,
        "fl_server_port": fl_server_port,
        "start_fl_job_threshold": start_fl_job_threshold,
        "start_fl_job_time_window": start_fl_job_time_window,
        "update_model_ratio": update_model_ratio,
        "update_model_time_window": update_model_time_window,
        "fl_name": fl_name,
        "fl_iteration_num": fl_iteration_num,
        "client_epoch_num": client_epoch_num,
        "client_batch_size": client_batch_size,
        "client_learning_rate": client_learning_rate,
        "worker_step_num_per_iteration": 65,
        "scheduler_manage_port": scheduler_manage_port,
        "config_file_path": config_file_path,
        "enable_ssl": enable_ssl,
        "encrypt_type": encrypt_type,
        "dp_eps": dp_eps,
        "dp_delta": dp_delta,
        "dp_norm_clip": dp_norm_clip,
        "sign_k": sign_k,
        "sign_eps": sign_eps,
        "sign_thr_ratio": sign_thr_ratio,
        "sign_global_lr": sign_global_lr,
        "sign_dim_out": sign_dim_out,
        "share_secrets_ratio": share_secrets_ratio,
        "cipher_time_window": cipher_time_window,
        "reconstruct_secrets_threshold": reconstruct_secrets_threshold,
        "root_first_ca_path": root_first_ca_path,
        "root_second_ca_path": root_second_ca_path,
        "pki_verify": pki_verify,
        "equip_crl_path": equip_crl_path,
        "replay_attack_time_diff": replay_attack_time_diff,
        "http_url_prefix": http_url_prefix,
        "global_iteration_time_window": global_iteration_time_window,
        "checkpoint_dir": checkpoint_dir,
        "client_password": "123456",
        "server_password": "123456"
    }
    for key in ctx:
        logging.info("ctx name: {}, value: {}".format(key, ctx[key]))
    context.set_context(mode=context.GRAPH_MODE, device_target=device_target, save_graphs=False)
    context.set_fl_context(**ctx)
    logging.info("Context setting is done!")

    """ load config file of yaml """
    server_config = load_config(server_config_path)
    net_model = init_model(server_config)

    cloud_task(net_model, init_checkpoint_dir=checkpoint_dir, input_shape=server_config["vae_params"]["input_shape"],
               latent_dim=server_config["vae_params"]["latent_dim"])